/* IBM_COPYRIGHT_BEGIN_TAG */
/*
 * Copyright IBM Corporation 2009. All rights reserved. The information contained
 * in these materials is provided for informational purposes only, and is provided AS IS
 * without warranty of any kind, express or implied. IBM shall not be responsible for any
 * damages arising out of the use of, or otherwise related to, these materials. Nothing
 * contained in these materials is intended to, nor shall have the effect of, creating any
 * warranties or representations from IBM or its suppliers or licensors, or altering the terms
 * and conditions of the applicable license agreement governing the use of IBM software.
 * References in these materials to IBM products, programs, or services do not imply that
 * they will be available in all countries in which IBM operates. Product release dates
 * and/or capabilities referenced in these materials may change at any time at IBM<92>s sole
 * discretion based on market opportunities or other factors, and are not intended to be a
 * commitment to future product or feature availability in any way. IBM, the IBM logo, AIX,
 * the AIX logo, and other IBM products and services are trademarks of the International
 * Business Machines Corporation, in the United States, other countries or both. Other
 * company, product, or service names may be trademarks or service marks of others.
 */
/* IBM_COPYRIGHT_END_TAG */

#include <pthread.h>
#include <stdio.h>
#include "Thread.h"
#include "Mutex.h"
#include "Machine.h"
#include "Resource.h"
#include <stdint.h>
#include "llrapi.h"
#include "rm_symbols.h"
#include "Scheduler.h"
#include "ClassRec.h"
#include "Log.h"
#include "MCM.h"
#include "Adapter.h"
#include "AdapterReq.h"

int
Machine::tasksAvailable(const Step* step) {
        const char* class_name = step->JobClass;
	ClassRec*   class_rec  = getClassRec(class_name);
	int         max_tasks;

	if (strcmp(StartdState, "Down") == 0) {
		log(1, "machine %s startd is down so it can not be used.\n", MachineName);
		return 0;
	}

	if (!_reconciled) {
		log(1, "machine %s is not reconciled so it can not be used.\n", MachineName);
		return 0;
	}

	log(2, "machine %s::tasks[%d/%d], class %s::tasks[%d/%d]\n",
	    MachineName, tasksRunning(), MaxTasks,
	    class_name, class_rec->tasksRunning(), class_rec->MaxTasks);
	max_tasks = MaxTasks-tasksRunning();
	max_tasks = std::min(max_tasks, class_rec->MaxTasks-class_rec->tasksRunning());

	AdapterReq* adapter_req = NULL;
	if (step->adapter_reqs.size() > 0) {
		adapter_req = step->adapter_reqs[0];
		bool isUS  = ((llr_network_mode_t)adapter_req->Mode) == LLR_US_MODE;
		bool isEN0 = (strcmp(adapter_req->TypeName, "en0") == 0);

		// assume sn_all, so must use all switch adapters
		// or, en0, so do not need to schedule any adapters
		for (int i=0; i<adapters.size(); i++) {
			if (((llr_network_type)adapters[i]->NetworkType) == LLR_SWITCH_NETWORK && !isEN0) {
				int total_windows = adapters[i]->TotalWindowCount;
				int used_windows  = adapters[i]->windowsInUse();
				log(2, "adapter_name=%s windows=[%d/%d], exclusive=%d\n",
				    adapters[i]->Name, used_windows, total_windows, adapters[i]->exclusiveUse());
				if (adapter_req->Usage == 1 && used_windows > 0) { // 1 = NOT_SHARED
					// since there is a job running, this adapter can't be used
					log(2, "step requested exclusive use of adapter, but it is used by another step.\n");
					max_tasks = 0;
				} else if (isUS) {
					// each task consumes adapter_req->Instances windows, but only if step requests US Mode 
					max_tasks = std::min(max_tasks, (total_windows-used_windows)/adapter_req->Instances);
				}

				if (adapters[i]->exclusiveUse()) {
					log(2, "adapter is being used exclusively, thus it is not available to this step.\n");
					max_tasks = 0;
				}
			}
		}
	}

	return std::max(max_tasks, 0);
}

Adapter*
Machine::getAdapter(const char* adapter_name) {
	Adapter* adapter = NULL;
	for (int i=0; adapter == NULL && i<adapters.size(); i++) {
		if (strcmp(adapter_name, adapters[i]->Name) == 0) {
			adapter = adapters[i];
		}
	}

	if (adapter == NULL) {
		adapter = new Adapter(adapter_name);
		adapter->get_ref();
		adapters.push_back(adapter);
	}

	return adapter;
}

void
Machine::clear_adapters() {
	for (int i=0; i<adapters.size(); i++) {
		adapters[i]->rel_ref();
	}

	adapters.clear();
}

ClassRec*
Machine::getClassRec(const char* class_name) {
	ClassRec* class_rec = NULL;
	for (int i=0; class_rec == NULL && i<class_recs.size(); i++) {
		if (strcmp(class_name, class_recs[i]->ClassName) == 0) {
			class_rec = class_recs[i];
		}
	}

	if (class_rec == NULL) {
		class_rec = new ClassRec(class_name);
		class_rec->get_ref();
		class_recs.push_back(class_rec);
	}

	return class_rec;
}

void
Machine::clear_class_recs() {
	for (int i=0; i<class_recs.size(); i++) {
		class_recs[i]->rel_ref();
	}

	class_recs.clear();
}

Resource*
Machine::getResource(const char* resource_name) {
	Resource* resource = NULL;
	for (int i=0; resource == NULL && i<resources.size(); i++) {
		if (strcmp(resource_name, resources[i]->Name) == 0) {
			resource = resources[i];
		}
	}

	if (resource == NULL) {
		resource = new Resource(resource_name);
		resource->get_ref();
		resources.push_back(resource);
	}

	return resource;
}

void
Machine::clear_resources() {
	for (int i=0; i<class_recs.size(); i++) {
		class_recs[i]->rel_ref();
	}

	class_recs.clear();
}

MCM*
Machine::getMCM(int mcm_id) {
	MCM* mcm = NULL;
	for (int i=0; mcm == NULL && i<mcms.size(); i++) {
		if (mcm_id == mcms[i]->ID) {
			mcm = mcms[i];
		}
	}

	if (mcm == NULL) {
		mcm = new MCM(mcm_id);
		mcm->get_ref();
		mcms.push_back(mcm);
	}

	return mcm;
}

void
Machine::clear_mcms() {
	for (int i=0; i<mcms.size(); i++) {
		mcms[i]->rel_ref();
	}

	mcms.clear();
}

bool
Machine::updateData(void* llr_machine, llr_event_type_t event_type) {
	Scheduler&     s   = Scheduler::getInstance();
	llr_element_t* err = NULL;
	int            rc  = 0;

	if (event_type == LLR_ALL_MACHINES_EVENT ||
	    event_type == LLR_MACHINE_UPDATE_EVENT) {

		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineDisk, Disk, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineDisk64, Disk64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineVirtualMemory, VirtualMemory, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineVirtualMemory64, VirtualMemory64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineFreeRealMemory, FreeRealMemory, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineFreeRealMemory64, FreeRealMemory64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesScanned, PagesScanned, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesScanned64, PagesScanned64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesFreed, PagesFreed, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesFreed64, PagesFreed64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesPagedIn, PagesPagedIn, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesPagedIn64, PagesPagedIn64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesPagedOut, PagesPagedOut, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePagesPagedOut64, PagesPagedOut64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineLargePageCount64, LargePageCount64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineLoadAverage, LoadAverage, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineStartdState, StartdState, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineStartdRunningJobs, StartdRunningJobs, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineKbddIdle, KbddIdle, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineSMTState, SMTState, err);
	}

	if (event_type == LLR_ALL_MACHINES_EVENT) {

		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineArchitecture, Architecture, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineTimeStamp, TimeStamp, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineRealMemory, RealMemory, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineRealMemory64, RealMemory64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineSpeed, Speed, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineMaxTasks, MaxTasks, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePrestartedStarters, PrestartedStarters, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineMachineMode, MachineMode, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachinePoolList, PoolList, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachinePoolListSize, PoolListSize, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineScheddState, ScheddState, err);
		GET_DATA_PP(rc, s.rm_handle, llr_machine, LLR_MachineConfiguredClassList, ConfiguredClassList, err);
		GET_DATA_PP(rc, s.rm_handle, llr_machine, LLR_MachineFeatureList, FeatureList, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineOperatingSystem, OperatingSystem, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineCPUs, CPUs, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineCPUList, CPUList, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineScheddRunningJobs, ScheddRunningJobs, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineScheddTotalJobs, ScheddTotalJobs, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineStartExpr, StartExpr, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineSuspendExpr, SuspendExpr, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineContinueExpr, ContinueExpr, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineVacateExpr, VacateExpr, err);
		GET_DATA_P(rc, s.rm_handle, llr_machine, LLR_MachineKillExpr, KillExpr, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineLargePageSize64, LargePageSize64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineLargePageFree64, LargePageFree64, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineConfigTimeStamp, ConfigTimeStamp, err);
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineRSetSupport, RSetSupport, err);

		llr_data_list_t llr_resource_list;
		void* llr_resource = NULL;
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineGetResourceList, llr_resource_list, err);
		GET_DATA(rc, s.rm_handle, &llr_resource_list, LLR_MachineGetFirstResource, llr_resource, err);
		while (llr_resource != NULL) {
			err = NULL;
			char* resource_name = NULL;
			GET_DATA_P(rc, s.rm_handle, llr_resource, LLR_ResourceName, resource_name, err);
			if (!err) {
				Resource* resource = getResource(resource_name);
				resource->updateData(llr_resource);
			}

			FREE_IF_SET(resource_name);

			GET_DATA(rc, s.rm_handle, &llr_resource_list, LLR_MachineGetNextResource, llr_resource, err);
		}

		llr_data_list_t llr_mcm_list;
		void* llr_mcm = NULL;
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineGetMCMList, llr_mcm_list, err);
		GET_DATA(rc, s.rm_handle, &llr_mcm_list, LLR_MachineGetFirstMCM, llr_mcm, err);
		while (llr_mcm != NULL) {
			err = NULL;
			int mcm_id = 0;
			GET_DATA(rc, s.rm_handle, llr_mcm, LLR_MCMID, mcm_id, err);
			if (!err) {
				MCM* mcm = getMCM(mcm_id);
				mcm->updateData(llr_mcm);
			}

			GET_DATA(rc, s.rm_handle, &llr_mcm_list, LLR_MachineGetNextMCM, llr_mcm, err);
		}

		char* class_name = NULL;
		int   class_init = 0;
		for (int i=0; ConfiguredClassList && ConfiguredClassList[i]; i++) {
			class_init++;
			if (class_name == NULL || strcmp(ConfiguredClassList[i], class_name) != 0) {
				if (class_name != NULL) {
					ClassRec* class_rec = getClassRec(class_name);
					class_rec->MaxTasks = class_init;
				}
				
				class_name = ConfiguredClassList[i];
				class_init = 0;
			}
		}
	}

	if (event_type == LLR_ALL_MACHINES_EVENT ||
	    event_type == LLR_MACHINE_UPDATE_ADAPTER_EVENT) {

		llr_data_list_t llr_adapter_list;
		void* llr_adapter = NULL;
		GET_DATA(rc, s.rm_handle, llr_machine, LLR_MachineGetAdapterList, llr_adapter_list, err);
		GET_DATA(rc, s.rm_handle, &llr_adapter_list, LLR_MachineGetFirstAdapter, llr_adapter, err);
		while (llr_adapter != NULL) {
			err = NULL;
			char* adapter_name = NULL;
			GET_DATA_P(rc, s.rm_handle, llr_adapter, LLR_AdapterName, adapter_name, err);
			if (!err) {
				Adapter* adapter = getAdapter(adapter_name);
				adapter->updateData(llr_adapter, event_type);
			}

			FREE_IF_SET(adapter_name);

			GET_DATA(rc, s.rm_handle, &llr_adapter_list, LLR_MachineGetNextAdapter, llr_adapter, err);
		}
	}

	display_machine_data(event_type);
	return (rc == 0);
}

void
Machine::display_machine_data(llr_event_type_t event_type) {

	log(1, "MACHINE MachineName:             %s\n",   MachineName);

	if (event_type == LLR_ALL_MACHINES_EVENT) {

		log(2, "MACHINE Architecture:            %s\n",   Architecture);
		log(2, "MACHINE TimeStamp:               %d\n",   TimeStamp);
		log(2, "MACHINE RealMemory:              %d\n",   RealMemory);
		log(2, "MACHINE RealMemory64:            %lld\n", RealMemory64);
		log(2, "MACHINE Speed:                   %d\n",   Speed);
		log(2, "MACHINE MaxTasks:                %d*\n",  MaxTasks);
		log(2, "MACHINE PrestartedStarters:      %d\n",   PrestartedStarters);
		log(2, "MACHINE MachineMode:             %s\n",   MachineMode);
		//log(2, "MACHINE PoolList:                \n", PoolList);
		log(2, "MACHINE PoolListSize:            %d\n",   PoolListSize);
		log(2, "MACHINE ScheddState:             %d\n",   ScheddState);
		for (int i=0; i<class_recs.size(); i++) {
			log(2, "MACHINE Class:                   %s(%d)\n", class_recs[i]->ClassName, class_recs[i]->MaxTasks);
		}

		//log(2, "MACHINE FeatureList:         \n", FeatureList);
		log(2, "MACHINE OperatingSystem:         %s\n",   OperatingSystem);
		log(2, "MACHINE CPUs:                    %d\n",   CPUs);
		for (int i=0; i<CPUs; i++) {
			log(2, "MACHINE CPU[%d]:                 %d\n", i, CPUList[i]);
		}
		log(2, "MACHINE ScheddRunningJobs:       %d\n",   ScheddRunningJobs);
		log(2, "MACHINE ScheddTotalJobs:         %d\n",   ScheddTotalJobs);
		log(2, "MACHINE StartExpr:               %s\n",   StartExpr);
		log(2, "MACHINE SuspendExpr:             %s\n",   SuspendExpr);
		log(2, "MACHINE ContinueExpr:            %s\n",   ContinueExpr);
		log(2, "MACHINE VacateExpr:              %s\n",   VacateExpr);
		log(2, "MACHINE KillExpr:                %s\n",   KillExpr);
		log(2, "MACHINE LargePageSize64:         %lld\n", LargePageSize64);
		log(2, "MACHINE LargePageFree64:         %lld\n", LargePageFree64);
		log(2, "MACHINE ConfigTimeStamp:         %d\n",   ConfigTimeStamp);
		log(2, "MACHINE RSetSupport:             %d\n",   RSetSupport);
	}

        if (event_type == LLR_ALL_MACHINES_EVENT ||
            event_type == LLR_MACHINE_UPDATE_EVENT) {

		log(2, "MACHINE Disk:                    %d\n",   Disk);
		log(2, "MACHINE Disk64:                  %lld\n", Disk64);
		log(2, "MACHINE VirtualMemory:           %d\n",   VirtualMemory);
		log(2, "MACHINE VirtualMemory64:         %lld\n", VirtualMemory64);
		log(2, "MACHINE FreeRealMemory:          %d\n",   FreeRealMemory);
		log(2, "MACHINE FreeRealMemory64:        %lld\n", FreeRealMemory64);
		log(2, "MACHINE PagesScanned:            %d\n",   PagesScanned);
		log(2, "MACHINE PagesScanned64:          %lld\n", PagesScanned64);
		log(2, "MACHINE PagesFreed:              %d\n",   PagesFreed);
		log(2, "MACHINE PagesFreed64:            %lld\n", PagesFreed64);
		log(2, "MACHINE PagesPagedIn:            %d\n",   PagesPagedIn);
		log(2, "MACHINE PagesPagedIn64:          %lld\n", PagesPagedIn64);
		log(2, "MACHINE PagesPagedOut:           %d\n",   PagesPagedOut);
		log(2, "MACHINE PagesPagedOut64:         %lld\n", PagesPagedOut64);
		log(2, "MACHINE LargePageCount64:        %lld\n", LargePageCount64);
		log(2, "MACHINE LoadAverage:             %d\n",   LoadAverage);
		log(2, "MACHINE StartdState:             %s\n",   StartdState);
		log(2, "MACHINE StartdRunningJobs:       %d\n",   StartdRunningJobs);
		log(2, "MACHINE KbddIdle:                %d\n",   KbddIdle);
		log(2, "MACHINE SMTState:                %d\n",   SMTState);
	}
}



Machine::Machine(char* machine_name) {
	MachineName         = strdup(machine_name);
	Architecture        = NULL;
	Disk                = 0;
	Disk64              = 0;
	TimeStamp           = 0;
	VirtualMemory       = 0;
	VirtualMemory64     = 0;
	RealMemory          = 0;
	RealMemory64        = 0;
	Speed               = 0;
	MaxTasks            = 0;
	PrestartedStarters  = 0;
	MachineMode         = NULL;
	PoolList            = NULL;
	PoolListSize        = 0;
	ScheddState         = 0;
	ConfiguredClassList = NULL;
	FeatureList         = NULL;
	OperatingSystem     = NULL;
	LoadAverage         = 0;
	FreeRealMemory      = 0;
	FreeRealMemory64    = 0;
	PagesScanned        = 0;
	PagesScanned64      = 0;
	PagesFreed          = 0;
	PagesFreed64        = 0;
	PagesPagedIn        = 0;
	PagesPagedIn64      = 0;
	PagesPagedOut       = 0;
	PagesPagedOut64     = 0;
	CPUs                = 0;
	CPUList             = NULL;
	StartdState         = NULL;
	ScheddRunningJobs   = 0;
	ScheddTotalJobs     = 0;
	StartdRunningJobs   = 0;
	KbddIdle            = 0;
	StartExpr           = NULL;
	SuspendExpr         = NULL;
	ContinueExpr        = NULL;
	VacateExpr          = NULL;
	KillExpr            = NULL;
	LargePageSize64     = 0;
	LargePageCount64    = 0;
	LargePageFree64     = 0;
	ConfigTimeStamp     = 0;
	RSetSupport         = 0;
	SMTState            = 0;

	_tasks_running      = 0;
	_steps_running      = 0;
	_reconciled         = false;
}

Machine::~Machine() {
	clear_adapters();
	clear_class_recs();
	clear_resources();
	clear_mcms();

	FREE_IF_SET(MachineName);
	FREE_IF_SET(Architecture);
	FREE_IF_SET(MachineMode);
	FREE_IF_SET(PoolList);
	FREE_IF_SET_ARR(ConfiguredClassList);
	FREE_IF_SET_ARR(FeatureList);
	FREE_IF_SET(OperatingSystem);
	FREE_IF_SET(CPUList);
	FREE_IF_SET(StartdState);
	FREE_IF_SET(StartExpr);
	FREE_IF_SET(SuspendExpr);
	FREE_IF_SET(ContinueExpr);
	FREE_IF_SET(VacateExpr);
	FREE_IF_SET(KillExpr);
}
