/* IBM_COPYRIGHT_BEGIN_TAG */
/*
 * Copyright IBM Corporation 2009. All rights reserved. The information contained
 * in these materials is provided for informational purposes only, and is provided AS IS
 * without warranty of any kind, express or implied. IBM shall not be responsible for any
 * damages arising out of the use of, or otherwise related to, these materials. Nothing
 * contained in these materials is intended to, nor shall have the effect of, creating any
 * warranties or representations from IBM or its suppliers or licensors, or altering the terms
 * and conditions of the applicable license agreement governing the use of IBM software.
 * References in these materials to IBM products, programs, or services do not imply that
 * they will be available in all countries in which IBM operates. Product release dates
 * and/or capabilities referenced in these materials may change at any time at IBM<92>s sole
 * discretion based on market opportunities or other factors, and are not intended to be a
 * commitment to future product or feature availability in any way. IBM, the IBM logo, AIX,
 * the AIX logo, and other IBM products and services are trademarks of the International
 * Business Machines Corporation, in the United States, other countries or both. Other
 * company, product, or service names may be trademarks or service marks of others.
 */
/* IBM_COPYRIGHT_END_TAG */

#include <stdlib.h>
#include <stdint.h>
#include "llrapi.h"
#include "MainThread.h"
#include "RMResetThread.h"
#include "EventListenerThread.h"
#include "SchedulingThread.h"
#include "Scheduler.h"
#include "Step.h"
#include "Log.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include "rm_symbols.h"

llr_init_resmgr_t           llr_init_resmgr_p;
llr_free_resmgr_t           llr_free_resmgr_p;
llr_register_for_events_t   llr_register_for_events_p;
llr_unregister_for_events_t llr_unregister_for_events_p;
llr_read_events_t           llr_read_events_p;
llr_get_event_t             llr_get_event_p;
llr_get_event_fd_t          llr_get_event_fd_p;
llr_free_event_t            llr_free_event_p;
llr_get_data_t              llr_get_data_p;
llr_start_job_step_t        llr_start_job_step_p;
llr_delete_job_t            llr_delete_job_p;
llr_error_t                 llr_error_p;

int        Scheduler::SCHEDULING_INTERVAL = 30*1000;
char*      Scheduler::SCHEDULER_LOG       = NULL;
int        Scheduler::DEBUG_LEVEL         = 1;
Scheduler* Scheduler::_instance           = NULL;

void
Scheduler::addStep(Step* step) {
	step->get_ref();
	steps.push_back(step);
	if (steps.size() >= 2) {
		int step_idx = steps.size() - 1;
		while (step_idx > 0 && steps[step_idx]->comparePriority(*(steps[step_idx-1])) < 0) {
			std::swap(steps[step_idx], steps[step_idx-1]);
			--step_idx;
		}
	}
}

Step*
Scheduler::findStep(char* step_id) {
	Step* step = NULL;
	steps_lock.lock();
	for (int i=0; step == NULL && i<steps.size(); i++) {
		steps[i]->lock.lock();
		if (strcmp(step_id, steps[i]->StepID) == 0) {
			step = steps[i];
		}
		steps[i]->lock.unlock();
	}

	steps_lock.unlock();

	return step;
}

void
Scheduler::deleteStep(char* step_id) {
	Step* step = NULL;
	steps_lock.lock();
	int i=0;
	for (; step == NULL && i<steps.size(); i++) {
		steps[i]->lock.lock();
		if (strcmp(step_id, steps[i]->StepID) == 0) {
			step = steps[i];
		}
		steps[i]->lock.unlock();
	}

	if (step) {
		step->rel_ref();
		step = NULL;
		for (int j=i; j<steps.size()-1; j++) {
			steps[j] = steps[j+1];
		}

		steps.pop_back();
	}

	steps_lock.unlock();
}

Job*
Scheduler::getJob(char* job_id) {
	Job* job = NULL;
	jobs_lock.lock();
	for (int i=0; job == NULL && i<jobs.size(); i++) {
		jobs[i]->lock.lock();
		if (strcmp(job_id, jobs[i]->JobID) == 0) {
			job = jobs[i];
		}
		jobs[i]->lock.unlock();
	}

	if (job == NULL) {
		job = new Job(job_id);
		job->get_ref();
		jobs.push_back(job);
	}
	jobs_lock.unlock();

	return job;
}

void
Scheduler::deleteJob(char* job_id) {
	Job* job = NULL;
	jobs_lock.lock();
	int i=0;
	for (; i<jobs.size(); i++) {
		jobs[i]->lock.lock();
		if (strcmp(job_id, jobs[i]->JobID) == 0) {
			job = jobs[i];
		}
		jobs[i]->lock.unlock();
		if (job) break;
	}

	if (job) {
		log(2, "deleteJob() successfully deleted job %s\n", job_id);
		if (i != jobs.size()-1) {
			jobs[i] = jobs.back();
		}
		jobs.pop_back();
		job->rel_ref();
		job = NULL;
	} else {
		log(0, "deleteJob() ERROR: job %s is not found in internal data structures\n", job_id);
	}
	jobs_lock.unlock();
}

Machine*
Scheduler::getMachine(char* machine_name) {
	Machine* machine = NULL;
	machines_lock.lock();
	for (int i=0; machine == NULL && i<machines.size(); i++) {
		machines[i]->lock.lock();
		if (strcmp(machine_name, machines[i]->MachineName) == 0) {
			machine = machines[i];
		}
		machines[i]->lock.unlock();
	}

	if (machine == NULL) {
		machine = new Machine(machine_name);
		machine->get_ref();
		machines.push_back(machine);
	}

	machines_lock.unlock();
	return machine;
}

void initialize_configuration() {
	Scheduler::SCHEDULER_LOG = getenv("SCHEDULER_LOG");

	char* temp = getenv("SCHEDULING_INTERVAL");
	if (temp != NULL) {
		Scheduler::SCHEDULING_INTERVAL = atoi(temp) * 1000;
	}

	temp = getenv("DEBUG_LEVEL");
	if (temp != NULL) {
		int level = atoi(temp);
		if (level >= 0 && level <= 6) {
			Scheduler::DEBUG_LEVEL = level;
		}
	}

	printf("LoadLeveler Sample Scheduler Starting up\n");
	printf("Settings:\n");
	printf("SCHEDULING_INTERVAL = %ds (time between each consecuitive scheduling cycle)\n",
	       Scheduler::SCHEDULING_INTERVAL/1000);
	printf("SCHEDULER_LOG       = %s\n", Scheduler::SCHEDULER_LOG);
	printf("DEBUG_LEVEL         = %d\n", Scheduler::DEBUG_LEVEL);
}

#define LOAD_SYMBOL(sym_t, sym_p, sym_name) {\
	sym_p = (sym_t)dlsym(handle, sym_name);\
	if (sym_p == NULL) {\
		log(0, "Sym %s failed: %s\n", sym_name, dlerror());\
		exit(-1);\
	} else {\
		log(2, "Sym %s successfull\n", sym_name);\
	}\
}

void load_symbols() {
	void* handle = NULL;

	char* libllrapi_path = "";

#ifdef __linux__
	libllrapi_path = "/usr/lib64/libllrapi.so";
	handle = (void*)dlopen(libllrapi_path, RTLD_LAZY);
#endif

#ifdef __aix__
	libllrapi_path = "/usr/lib/libllrapi.a(shr.o)";
	handle = (void*)dlopen(libllrapi_path, RTLD_LAZY|RTLD_MEMBER);
#endif

	if (handle == NULL) {
		perror("dlopen:libllrapi");
		log(0, "Load of \"%s\" failed: %s\n", libllrapi_path, dlerror());
		exit(-1);
	} else {
		log(2, "Load successfull\n");
	}

	LOAD_SYMBOL(llr_init_resmgr_t,           llr_init_resmgr_p,           LLR_INIT_RESMGR);
	LOAD_SYMBOL(llr_free_resmgr_t,           llr_free_resmgr_p,           LLR_FREE_RESMGR);
	LOAD_SYMBOL(llr_register_for_events_t,   llr_register_for_events_p,   LLR_REGISTER_FOR_EVENTS);
	LOAD_SYMBOL(llr_unregister_for_events_t, llr_unregister_for_events_p, LLR_UNREGISTER_FOR_EVENTS);
	LOAD_SYMBOL(llr_read_events_t,           llr_read_events_p,           LLR_READ_EVENTS);
	LOAD_SYMBOL(llr_get_event_t,             llr_get_event_p,             LLR_GET_EVENT);
	LOAD_SYMBOL(llr_get_event_fd_t,          llr_get_event_fd_p,          LLR_GET_EVENT_FD);
	LOAD_SYMBOL(llr_free_event_t,            llr_free_event_p,            LLR_FREE_EVENT);
	LOAD_SYMBOL(llr_get_data_t,              llr_get_data_p,              LLR_GET_DATA);
	LOAD_SYMBOL(llr_start_job_step_t,        llr_start_job_step_p,        LLR_START_JOB_STEP);
	LOAD_SYMBOL(llr_delete_job_t,            llr_delete_job_p,            LLR_DELETE_JOB);
	LOAD_SYMBOL(llr_error_t,                 llr_error_p,                 LLR_ERROR);
}


bool
Scheduler::register_with_rm() {
	llr_element_t* err_object = NULL;
	int rc = 0; 

	llr_resmgr_handle_t*&     rm_handle = Scheduler::getInstance().rm_handle;
	llr_event_registration_t& reg_data  = Scheduler::getInstance().reg_data;

	log(2, "before llr_init_resmgr\n");
	rc = llr_init_resmgr_p(LLR_API_VERSION, &rm_handle, &err_object);
	log(1, "return from llr_init_resmgr: rc = %d handle = %p\n", rc, rm_handle);

	reg_data.port = 0;
	reg_data.scheduler_id = "sample_scheduler";

	if (rm_handle != NULL && err_object == NULL) {
		// Register for event
		rc = llr_register_for_events_p(rm_handle, &reg_data, &err_object);
		log(1, "return from llr_register_for_events: rc = %d \n", rc);
	}

	if (err_object) {
		llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
	} else {
		// Get the file descriptor for the listen socket
		int fd = llr_get_event_fd_p(rm_handle);
		log(1, "Event file descriptor = %d\n", fd);
	}

	return (rm_handle != NULL && err_object == NULL && rc == 0);
}

bool
Scheduler::unregister_with_rm() {
	llr_element_t *err_object = NULL;
	int rc; 

	llr_resmgr_handle_t*&     rm_handle = Scheduler::getInstance().rm_handle;
	llr_event_registration_t& reg_data  = Scheduler::getInstance().reg_data;

	// Unregister for event
	log(2, "before llr_unregister_for_events\n");
	rc = llr_unregister_for_events_p(rm_handle, &reg_data, &err_object);
	log(1, "return from llr_unregister_for_events: rc = %d \n", rc);
	if (err_object) {
		llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
	}

	// File descriptor should be -1
	rc = llr_get_event_fd_p(rm_handle);
	log(2, "Event file descriptor = %d\n",rc);

	err_object = NULL;
	rc = llr_free_resmgr_p(&rm_handle, &err_object);
	log(2, "return from llr_free_resmgr: rc = %d handle = %p\n", rc, rm_handle);

	return (rm_handle == NULL && err_object == NULL);
}

void
Scheduler::RMReset() {
	_rm_reset_in_progress_lock.lock();
	if (!_rm_reset_in_progress) {
		log(1, "Scheduler::RMReset(): Starting RM_RESET_THREAD and marking reset_in_progress=TRUE\n");
		_rm_reset_in_progress = true;
		RMResetThread* rm_reset_thread = new RMResetThread();
		rm_reset_thread->startThread();
	} else {
		log(1, "Scheduler::RMReset(): No need to start RM_RESET_THREAD since reset_in_progress=TRUE\n");
	}
	_rm_reset_in_progress_lock.unlock();
}

void
Scheduler::RMMarkDown() {
	// _rm_state_lock is used to make sure that reset and mark_down
	// are not taking place at the same time
	_rm_state_lock.lock();
	log(0, "Scheduler::RMMarkDown(): Marking RM as DOWN\n");
	if (RMisUP()) {
		log(0, "RM_RESET_THREAD: RM is UP, so need to first unregister from the RM\n");
		bool success = unregister_with_rm();
		RMisUP(false);
		if (success) {
			log(0, "RM_RESET_THREAD: successfully unregistered from the RM\n");
		} else {
			log(0, "RM_RESET_THREAD: could not unregister from the RM\n");
		}
	} else {
		log(0, "RM_RESET_THREAD: RM is not UP, no need to unregister from the RM\n");
	}
	_rm_state_lock.unlock();
}

void
Scheduler::eventListenerThread(EventListenerThread* event_listener_thread) {
	log(2, "EVENT_LISTENER_THREAD OLD = 0x%x\n", _event_listener_thread);
	if (_event_listener_thread) {
		_event_listener_thread->rel_ref();
	}

	_event_listener_thread = event_listener_thread;

	if (_event_listener_thread) {
		_event_listener_thread->get_ref();
	}
	log(1, "EVENT_LISTENER_THREAD NEW = 0x%x\n", _event_listener_thread);
}

void
Scheduler::schedulingThread(SchedulingThread* scheduling_thread) {
	if (_scheduling_thread) {
		_scheduling_thread->rel_ref();
	}

	_scheduling_thread = scheduling_thread;

	if (_scheduling_thread) {
		_scheduling_thread->get_ref();
	}
}


int main(int argc, char** argv, char** envp) {
	initialize_configuration();
	initialize_log();

	load_symbols();

	MainThread* main_thread = new MainThread();

	main_thread->startThread();
	main_thread->join();
	printf("finish\n");

	close_log();
}
