/* IBM_COPYRIGHT_BEGIN_TAG */
/*
 * Copyright IBM Corporation 2009. All rights reserved. The information contained
 * in these materials is provided for informational purposes only, and is provided AS IS
 * without warranty of any kind, express or implied. IBM shall not be responsible for any
 * damages arising out of the use of, or otherwise related to, these materials. Nothing
 * contained in these materials is intended to, nor shall have the effect of, creating any
 * warranties or representations from IBM or its suppliers or licensors, or altering the terms
 * and conditions of the applicable license agreement governing the use of IBM software.
 * References in these materials to IBM products, programs, or services do not imply that
 * they will be available in all countries in which IBM operates. Product release dates
 * and/or capabilities referenced in these materials may change at any time at IBM<92>s sole
 * discretion based on market opportunities or other factors, and are not intended to be a
 * commitment to future product or feature availability in any way. IBM, the IBM logo, AIX,
 * the AIX logo, and other IBM products and services are trademarks of the International
 * Business Machines Corporation, in the United States, other countries or both. Other
 * company, product, or service names may be trademarks or service marks of others.
 */
/* IBM_COPYRIGHT_END_TAG */

#include <stdlib.h>
#include <pthread.h>
#include "SchedulingThread.h"
#include "Scheduler.h"
#include "Step.h"
#include "stdio.h"
#include "Log.h"
#include "rm_symbols.h"
#include "Node.h"
#include "AdapterReq.h"

void
SchedulingThread::process_events() {
	Scheduler& scheduler = Scheduler::getInstance();
	std::vector<llr_event_t*> events;
	llr_element_t *err_object;
	int rc;

	scheduler.event_queue_lock.lock();
	while (!scheduler.event_queue.empty()) {
		events.push_back(scheduler.event_queue.front());
		scheduler.event_queue.pop_front();
	}
	scheduler.event_queue_lock.unlock();

	for (int i=0; i<events.size(); i++) {
		llr_event_t* event        = events[i];

		log(1, "SCHEDULING_THREAD: Processing %s event\n", enum_to_string(event->event_type));

		switch (event->event_type) {
		case LLR_RESOURCE_MGR_DOWN_EVENT:
			scheduler.RMMarkDown();
			break;
		case LLR_RESOURCE_MGR_UP_EVENT:
			for (int i=0; i<scheduler.machines.size(); i++) {
				log(2, "Mark machine %s unreconciled\n",
				    scheduler.machines[i]->MachineName);
				scheduler.machines[i]->_reconciled = false;
			}
			scheduler.RMReset();
			break;
		case LLR_ALL_MACHINES_EVENT: {
			for (int i=0; i<scheduler.machines.size(); i++) {
				log(2, "Mark machine %s unreconciled\n",
				    scheduler.machines[i]->MachineName);
				scheduler.machines[i]->_reconciled = false;
			}
			llr_element_t** machine_list = (llr_element_t**)event->event_data;
			char*           machine_name = NULL;
			for (int j=0; machine_list[j] != NULL; j++) {
				err_object = NULL;
				rc = llr_get_data_p(scheduler.rm_handle, machine_list[j], LLR_MachineName, (void*)&machine_name, &err_object); 
				if (err_object) {
					llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
				} else {
					log(1, "Machine[%d] = %s (%p)\n", j, machine_name, machine_list[j]);
					Machine* mach = scheduler.getMachine(machine_name);
					mach->lock.lock();
					mach->updateData(machine_list[j], event->event_type);
					mach->lock.unlock();
				}

				if (machine_name != NULL) {
					free(machine_name);
					machine_name = NULL;
				}
			}
		}
			break;
		case LLR_ALL_JOBS_EVENT: {
			llr_element_t** job_list = (llr_element_t**)event->event_data;
			char*           job_id   = NULL;
			for (int j=0; job_list[j] != NULL; j++) {
				err_object = NULL;
				rc = llr_get_data_p(scheduler.rm_handle, job_list[j], LLR_JobID, (void*)&job_id, &err_object); 
				if (err_object) {
					llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
				} else {
					log(1, "Job[%d] = %s (%p)\n", j, job_id, job_list[j]);
					Job* job = scheduler.getJob(job_id);
					job->lock.lock();
					job->updateData(job_list[j]);
					for (int i=0; i<job->steps.size(); i++) {
						Step* step = job->steps[i];
						bool running_like = step->isRunningLike();
						bool consumes_res = step->consumesResources();
						if (running_like && !consumes_res) {
							decrease_resources(step);
						} else if (!running_like && consumes_res) {
							increase_resources(step);
						}
						log(1, "Step[%d] running_like = %s, consumes_resources = %s, State = %s\n",
						    i,
						    running_like ? "YES" : "NO",
						    consumes_res ? "YES" : "NO",
						    enum_to_string((llr_step_state_t)step->State));
					}
					job->lock.unlock();
				}

				if (job_id != NULL) {
					free(job_id);
					job_id = NULL;
				}
			}
		}
			break;
		case LLR_JOB_MGR_DOWN_EVENT:
			break;
		case LLR_JOB_MGR_UP_EVENT:
			break;
		case LLR_MACHINE_DOWN_EVENT:
			break;
		case LLR_MACHINE_UPDATE_EVENT: 
		case LLR_MACHINE_UPDATE_ADAPTER_EVENT: {
			llr_element_t* llr_machine = (llr_element_t*)event->event_data;
			char* machine_name = NULL;
			err_object = NULL;
			rc = llr_get_data_p(scheduler.rm_handle, llr_machine, LLR_MachineName, (void*)&machine_name, &err_object); 
			if (err_object) {
				llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
			} else {
				log(1, "Machine = %s\n", machine_name);
				Machine* mach = scheduler.getMachine(machine_name);
				mach->lock.lock();
				mach->updateData(llr_machine, event->event_type);
				mach->lock.unlock();
			}

			if (machine_name != NULL) {
				free(machine_name);
				machine_name = NULL;
			}
		}
			break;
		case LLR_MACHINE_CONFIG_EVENT:
			break;
		case LLR_JOB_STEP_STATE_CHANGE_EVENT: {
			llr_state_change_event_data_t* state_change = (llr_state_change_event_data_t*)event->event_data;
			Step* step = scheduler.findStep(state_change->job_step_id);
			if (step != NULL) {
				step->lock.lock();
				log(0, "Change state for %s, %s(cur)->%s(new)\n",
				    step->StepID,
				    enum_to_string((llr_step_state_t)step->State),
				    enum_to_string(state_change->job_step_state));	
				bool was_running_like = step->isRunningLike();
				bool consumed_resources = step->consumesResources();
				step->State = (int)state_change->job_step_state;
				bool is_running_like = step->isRunningLike();
				if ((was_running_like || consumed_resources) && !is_running_like) {
					increase_resources(step);
				}
				log(1, "Was running-like = %s, consumes-resources = %s, is running-like = %s\n",
				    was_running_like ? "YES" : "NO",
				    consumed_resources ? "YES" : "NO",
				    is_running_like ? "YES" : "NO");
				step->lock.unlock();
			} else {
				log(0, "ERROR: step %s is not found\n", state_change->job_step_id);
			}
		}
			break;
		case LLR_JOB_DELETE_EVENT: {
			char* job_id = (char*)event->event_data;
			log(0, "DELETE job %s\n", job_id);
			scheduler.deleteJob(job_id);
		}
			break;
		case LLR_JOB_ADD_EVENT: {
			llr_element_t** llr_job = (llr_element_t**)event->event_data;
			char*           job_id  = NULL;
			err_object = NULL;
			rc = llr_get_data_p(scheduler.rm_handle, llr_job, LLR_JobID, (void*)&job_id, &err_object); 
			if (err_object) {
				llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
			} else {
				log(1, "Job = %s (%p)\n", job_id, llr_job);
				Job* job = scheduler.getJob(job_id);
				job->lock.lock();
				job->updateData(llr_job);
				job->lock.unlock();
			}

			if (job_id != NULL) {
				free(job_id);
				job_id = NULL;
			}
			break;
		}
		default:
			log(0, "Received unrecognized event; event_type=%d\n", event->event_type);
			break;
		}

		err_object = NULL;

		// Free event data
		rc = llr_free_event_p(scheduler.rm_handle, &event, &err_object);
		log(2, "return from free_event: rc = %d \n", rc);
		if (err_object) {
			llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
		}
	}
}

void
SchedulingThread::scheduling_cycle() {
	Scheduler& scheduler = Scheduler::getInstance();

	log(0, "\n");
	log(0, "ENTER Scheduling Cycle\n");

	if (scheduler.RMisUP()) {
		log(2, "Resource Manager is UP\n");

		scheduler.steps_lock.lock();
		for (int i=0; i<scheduler.steps.size(); i++) {
			Step* step = scheduler.steps[i];

			step->lock.lock();
			if (step->State == LLR_STATE_IDLE && !step->consumesResources()) {
				schedule_step(step);
			}
			step->lock.unlock();
		}
		scheduler.steps_lock.unlock();
	} else {
		log(0, "Resource Manager is DOWN\n");
	}

	log(0, "EXIT Scheduling Cycle\n");
	log(0, "\n");
}

void
SchedulingThread::schedule_step(Step* step) {
	Scheduler& scheduler = Scheduler::getInstance();

	log(1, "schedule_step(%s)\n", step->StepID);

	if (step->nodes.size() != 1) {
		log(1, "skip step %s\n", step->StepID);
		log(1, "only know how to schedule steps with fixed tasks_per_node\n");
		return;
	}

	int machines_needed = step->nodes[0]->MaxInstances;
	int tasks_per_node  = step->nodes[0]->InitiatorCount;

	log(1, "nodes = %d, tasks_per_node = %d\n", machines_needed, tasks_per_node);

	AdapterReq* adapter_req = NULL;
	if (step->adapter_reqs.size() > 0) {
		adapter_req = step->adapter_reqs[0];
		if (strcmp(adapter_req->TypeName, "sn_all") != 0 &&
		    strcmp(adapter_req->TypeName, "eth0") != 0 &&
		    strcmp(adapter_req->TypeName, "en0") != 0) {
			log(1, "skip step %s\n", step->StepID);
			log(1, "step requests \"%s\" but only know how to schedule \"sn_all\", \"en0\" and \"eth0\"\n",
			    adapter_req->TypeName);
			return;
		}
	}

	std::vector<Machine*> machines;

	for (int i=0; i<scheduler.machines.size() && machines_needed>0; i++) {
		int tasks_available = scheduler.machines[i]->tasksAvailable(step);
		if (tasks_available >= tasks_per_node) {
			machines.push_back(scheduler.machines[i]);
			--machines_needed;
		}
	}

	if (machines_needed == 0) {
		log(1, "Found enough resources for step %s\n", step->StepID);

		assign_resources(step, machines);
		decrease_resources(step);
		start_step(step);

	} else {
		log(1, "Not enough resources available for step %s. Found %d out of %d required machines\n",
		    step->StepID, machines.size(), machines.size() + machines_needed);
	}
}

void
SchedulingThread::decrease_resources(Step* step) {
	step->consumesResources(true);
	log(1, "-R(%s)\n", step->StepID);
	int         tasks_per_node  = step->nodes[0]->InitiatorCount;
	const char* class_name      = step->JobClass;
	AdapterReq* adapter_req     = NULL;

	if (step->adapter_reqs.size() > 0) {
		adapter_req = step->adapter_reqs[0];
	}

	for (int i=0; i<step->machines.size(); i++) {
		Machine* machine = step->machines[i];

		int tasks_running = machine->tasksRunning();
		int steps_running = machine->stepsRunning();
		machine->tasksRunning(tasks_running + tasks_per_node);
		machine->stepsRunning(steps_running + 1);

		if (adapter_req && strcmp(adapter_req->TypeName, "sn_all") == 0) {
			bool isUS = ((llr_network_mode_t)adapter_req->Mode) == LLR_US_MODE;

			// consume on all switch adapters
			for (int j=0; j<machine->adapters.size(); j++) {
				Adapter* adapter = machine->adapters[j];
				if (((llr_network_type)adapter->NetworkType) == LLR_SWITCH_NETWORK) {
					if (isUS) {
						int windows_in_use = adapter->windowsInUse();
						adapter->windowsInUse(windows_in_use + (tasks_per_node * adapter_req->Instances));
					}
					adapter->exclusiveUse(adapter_req->Usage ? 1 : 0);
				}
			}
		}

		ClassRec* class_rec = machine->getClassRec(class_name);
		tasks_running = class_rec->tasksRunning();
		class_rec->tasksRunning(tasks_running + tasks_per_node);
	}
}

void
SchedulingThread::increase_resources(Step* step) {
	step->consumesResources(false);
	log(1, "+R(%s)\n", step->StepID);
	int         tasks_per_node  = step->nodes[0]->InitiatorCount;
	const char* class_name      = step->JobClass;
	AdapterReq* adapter_req     = NULL;

	if (step->adapter_reqs.size() > 0) {
		adapter_req = step->adapter_reqs[0];
	}

	for (int i=0; i<step->machines.size(); i++) {
		Machine* machine = step->machines[i];

		int tasks_running = machine->tasksRunning();
		int steps_running = machine->stepsRunning();
		machine->tasksRunning(tasks_running - tasks_per_node);
		machine->stepsRunning(steps_running - 1);

		if (adapter_req && strcmp(adapter_req->TypeName, "sn_all") == 0) {
			bool isUS = ((llr_network_mode_t)adapter_req->Mode) == LLR_US_MODE;

			// release on all switch adapters
			for (int j=0; j<machine->adapters.size(); j++) {
				Adapter* adapter = machine->adapters[j];
				if (((llr_network_type)adapter->NetworkType) == LLR_SWITCH_NETWORK) {
					if (isUS) {
						int windows_in_use = adapter->windowsInUse();
						adapter->windowsInUse(windows_in_use - (tasks_per_node * adapter_req->Instances));
					}
					adapter->exclusiveUse(0);
				}
			}
		}

		ClassRec* class_rec = machine->getClassRec(class_name);
		tasks_running = class_rec->tasksRunning();
		class_rec->tasksRunning(tasks_running - tasks_per_node);
	}
}

void
SchedulingThread::assign_resources(Step* step, std::vector<Machine*> machines) {
	step->clear_machines();
	for (int i=0; i<machines.size(); i++) {
		step->addMachine(machines[i]);
	}
}

void
SchedulingThread::start_step(Step* step) {
	if (step->machines.size() == 0) {
		log(0, "start_step(%s) ERROR: Machine list is empty.\n", step->StepID);
		return;
	}

	Scheduler& scheduler = Scheduler::getInstance();

	llr_job_step_resource_t resources;
	llr_job_step_opts_t     opts;
	llr_machine_data_t*     machine_list = new llr_machine_data_t[step->machines.size()];

	opts.flags = (llr_job_start_flags_t)0;
	resources.machine_count       = step->machines.size();
	resources.machine_list        = machine_list;
	resources.network_usage_count = 0;
	resources.network_usage_list  = NULL;
	resources.bg_partition        = NULL;

	int tasks_per_node = step->nodes[0]->InitiatorCount;

	AdapterReq* adapter_req     = NULL;
	if (step->adapter_reqs.size() > 0) {
		adapter_req = step->adapter_reqs[0];
	}

	if (adapter_req &&
	    (strcmp(adapter_req->TypeName, "sn_all") == 0 ||
	    (strcmp(adapter_req->TypeName, "en0") == 0) ||
	     strcmp(adapter_req->TypeName, "eth0") == 0)) {
		bool isUS     = ((llr_network_mode_t)adapter_req->Mode) == LLR_US_MODE;
		bool isEN0    = (strcmp(adapter_req->TypeName, "en0") == 0) || (strcmp(adapter_req->TypeName, "eth0") == 0);
		bool isSN_ALL = (strcmp(adapter_req->TypeName, "sn_all") == 0);
		resources.network_usage_count                  = 1; // for sn_all, 1 network_usage describes all networks
		resources.network_usage_list                   = new llr_network_usage_t[1];
		resources.network_usage_list[0].network_id     = isSN_ALL ? 0 : 0xFFFFFFFF; // for sn_all, network_id is 0
		resources.network_usage_list[0].instances      = adapter_req->Instances;
		resources.network_usage_list[0].window_count   = isUS ? 1 : 0;
		resources.network_usage_list[0].adapter_memory = 0;
		resources.network_usage_list[0].protocol       = adapter_req->Protocol;
		resources.network_usage_list[0].mode           = isUS  ? LLR_US_MODE : LLR_IP_MODE;
		resources.network_usage_list[0].network_type   = isEN0 ? LLR_ETHERNET_NETWORK : LLR_SWITCH_NETWORK;
	}
	int task_id = 0;
	for (int i=0; i<step->machines.size(); i++) {
		machine_list[i].machine_name = step->machines[i]->MachineName;
		machine_list[i].task_count   = tasks_per_node;
		machine_list[i].task_list    = new llr_task_data_t[tasks_per_node];
		for (int j=0; j<tasks_per_node; j++) {
			machine_list[i].task_list[j].task_id   = task_id;
			machine_list[i].task_list[j].cpu_count = 0;
			machine_list[i].task_list[j].cpu_list  = NULL;
			++task_id;
		}
	}

	llr_element_t* err_object = NULL;
	int rc = llr_start_job_step_p(scheduler.rm_handle, step->StepID, &resources, &opts, &err_object);

	if (err_object) {
		llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
	}

	log(1, "return from start_job_step: rc = %d handle = %p\n", rc, scheduler.rm_handle);

	// clean up resources data structure
	if (resources.network_usage_list) {
		delete[] resources.network_usage_list;
	}

	for (int i=0; i<step->machines.size(); i++) {
		delete[] machine_list[i].task_list;
	}

	delete[] machine_list;
}

void
SchedulingThread::reconcile_machines() {
	Scheduler& scheduler = Scheduler::getInstance();

	for (int i=0; i<scheduler.machines.size(); i++) {
		Machine* mach = scheduler.machines[i];
		mach->lock.lock();
		if (!mach->_reconciled) {
			if (mach->tasksRunning() == mach->StartdRunningJobs) {
				log(2, "Machine %s is reconciled\n", mach->MachineName);
				mach->_reconciled = true;
			} else {
				log(1, "Machine %s is not reconciled: tasks_running = %d, startd_running_jobs(tasks) = %d\n",
				    mach->MachineName, mach->tasksRunning(), mach->StartdRunningJobs);
			}
		}
		mach->lock.unlock();
	}
}

void
SchedulingThread::cleanup_jobs() {
	llr_element_t* err_object = NULL;
	Scheduler& scheduler = Scheduler::getInstance();

	scheduler.jobs_lock.lock();
	for (int i=0; i<scheduler.jobs.size(); i++) {
		Job* job = scheduler.jobs[i];
		bool all_steps_completed = true;
		for (int j=0; j<job->steps.size(); j++) {
			if (((llr_step_state_t)job->steps[j]->State) != LLR_STATE_COMPLETED &&
			    ((llr_step_state_t)job->steps[j]->State) != LLR_STATE_REMOVED) {
				all_steps_completed = false;
				break;
			}
		}
		if (all_steps_completed) {
			log(0, "All steps of job %s completed. Deleting the job.\n", job->JobID);
			llr_delete_job_p(scheduler.rm_handle, job->JobID, &err_object);
			if (err_object) {
				llr_error_p(&err_object, LLR_ERROR_PRINT_STDERR);
			}
		}
	}
	scheduler.jobs_lock.unlock();
}

void
SchedulingThread::signal() {
	signaled = true;
	cond->signal();
}

void
SchedulingThread::initialize() {
	if (cond == NULL) {
		cond = new Condition();
	}
}

SchedulingThread::~SchedulingThread() {
	if (cond != NULL) {
		delete cond;
	}
}

void
SchedulingThread::run() {
	log(0, "Starting SCHEDULING_THREAD\n");
	while (true) {
		process_events();
		reconcile_machines();
		cleanup_jobs();
		scheduling_cycle();
		log(0, "SCHEDULING_THREAD: going to sleep for %d seconds\n", Scheduler::SCHEDULING_INTERVAL/1000);
		bool expire = false;
		if (!signaled) {
			expire = cond->timed_wait(Scheduler::SCHEDULING_INTERVAL);
		}
		signaled = false;
		log(0, "SCHEDULING_THREAD: woke up for next scheduling cycle. Wait interrupted = %s\n", expire ? "false" : "true");
	}
}
